# --------------------------------------------------------------------
# jokettt project, v. 0.1
# by F. Piantini <francesco.piantini@gmail.com>
# ---
# Tic Tac Toe Console Player class definition
# --------------------------------------------------------------------
# pylint: disable=too-few-public-methods

"""Implementation of the Console TttPlayer:
    a player that gets the moves from console. This can be used to make
    able to play an human player using a minimal command line interface.
    The moves shall be entered using the format <row><column>, with
    <row> = A, B, C and <col> = 1, 2, 3.
    This class is derived from the TttPlayer base class
"""
from tttplayer import TttPlayer

class TttConsolePlayer(TttPlayer):
    """A Tic Tac Toe console player."""
    # no need to redefine the __init__ from TttPlayer
    #def __init__(self, piece):
    #    pass

    def move(self, board):
        """Method to make a move. Just read the move from console (stdin)"""
        # read move from Console
        while True:
            move = input("Move? ")
            if board.is_valid_move(move):
                return board.convert_movestring_to_indexes(move)
            print("Invalid move. Try again.")



# --------------------------------------------------------------------
# jokettt project, v. 0.1
# by F. Piantini <francesco.piantini@gmail.com>
# ---
# Tic Tac Toe "Learner" Player class definition
# --------------------------------------------------------------------
"""Implementation of a "learner" tic-tac-toe player, that improves
    his gameplay learning during games.
    The player made use of a value function that assign a value to
    every board state. Initially the value is 0 for any losing position,
    1 for every winning position and 0.5 for all the other positions.
    During games the values of the intermediates position are calibrated
    using the standard reinforcement learning formula:
        V(s) = V(s) + alpha * [ V(s') - V(s) ]
    This class is derived from the TttPlayer base class
"""
from random import shuffle

from tttplayer import TttPlayer

class TttLearnerPlayer(TttPlayer):
    """A Tic Tac Toe learner automatic player."""

    # --------------------------------------------------------------
    def __init__(self, piece, board, alpha=0.1, verbosity=0):
        """TttPlayer class constructor. Save the given piece,
            the alpha value and initializes Value vector."""
        print("verbosity level = ", verbosity)
        TttPlayer.__init__(self, piece, verbosity)
        self.__alpha = alpha
        self.__values = {}
        self.__best_value = -1000
        self.__best_x = None
        self.__best_y = None
        self.__best_zhash = -1
        zhash = board.get_zhash()
        score = board.evaluate(self.piece)
        if score > 0:
            # winning board...
            self.__values[zhash] = 1.0
        elif score < 0:
            #losing board
            self.__values[zhash] = 0.0
        else:
            # playable board
            self.__values[zhash] = 0.5
        self.__last_zhash = zhash

    # --------------------------------------------------------------
    def move(self, board):
        """Do a move using reinforcement learning algo"""
        return self.__find_rl_move(board)

    # --------------------------------------------------------------
    def learn_from_defeat(self, board):
        """Updates the value vector given a final lost position"""
        current_zhash = board.get_zhash()
        score = board.evaluate(self.piece) # this should be negative...
        if score < 0:            # so this check is useless...
            # defeat...
            self.__values[current_zhash] = 0.0
            self.__values[self.__last_zhash] += \
            self.__alpha * (self.__values[current_zhash] - \
                            self.__values[self.__last_zhash])
            self.log_info("LEARNED FROM DEFEAT... new value for last position: ",
                          self.__values[self.__last_zhash])

    # --------------------------------------------------------------
    def __find_rl_move(self, board):
        """Find a move that is considered the best depending on current knowledge"""
        current_zhash = board.get_zhash()
        if not current_zhash in self.__values:
            # the board status is not in values array:
            # this is the first time we encounter this position
            self.log_info("new board status encounted: init to 0.5")
            self.__values[current_zhash] = 0.5

        move_list = board.valid_moves()
        # interestingly, if we shuffle the possible moves before to select them,
        # the learning playing against a minimax player is slower
        shuffle(move_list)  # to add some variability to the play (...maybe)

        self.__best_value = -1000
        self.__best_x = None
        self.__best_y = None
        for move in move_list:
            if self.__analyze_move(move, board):
                # winning move found
                break

        # move selected... updates current zhash
        self.__values[current_zhash] += \
            self.__alpha * (self.__values[self.__best_zhash] - \
                            self.__values[current_zhash])
        self.__values[self.__last_zhash] += \
            self.__alpha * (self.__values[self.__best_zhash] - \
                            self.__values[self.__last_zhash])

        self.log_info("Position value updated. New value = ", self.__values[current_zhash])

        self.__last_zhash = self.__best_zhash
        return self.__best_x, self.__best_y

    # --------------------------------------------------------------
    def __analyze_move(self, move, board):
        """Analyze the move "move" given the current "board" status
            Returns True if the move is winning"""

        # do the move. The analyze_move() method returns the zhash
        # and the score of the board after the given move.
        # The score can be > 0 if the move is winning, or 0 if
        # the move does not cause a win.
        # Note that it is impossible that the score is < 0: it is
        # impossible to enter in lost position when it is our turn
        # to move
        zhash, score = board.analyze_move(move, self.piece)
        self.log_info("evaluating move: ", board.convert_move_to_movestring(move),
                      ", score = ", score)

        if score > 0:
            # we win! choose this move
            self.log_info("WINNING MOVE! Choose it")
            self.__values[zhash] = 1.0
        else:
            # neutral move... if the hash is not in dictionary
            # this is the first time we encounter this move:
            # initialize value
            self.log_info("NEUTRAL MOVE... analyzing it")
            if not zhash in self.__values:
                self.log_info("   - new board status encounted: init to 0.5")
                self.__values[zhash] = 0.5
            else:
                self.log_info("   - I know this move... value = ", self.__values[zhash])

        # It the value of the board after the move is better of values
        # seen until now, save the move data
        if self.__best_value < self.__values[zhash]:
            self.__best_zhash = zhash
            self.__best_value = self.__values[zhash]
            self.__best_x, self.__best_y = move
            self.log_info("   - move is selected as the new best choice - value = ",
                          self.__best_value)

        return score > 0



# --------------------------------------------------------------------
# jokettt project, v. 0.1
# by F. Piantini <francesco.piantini@gmail.com>
# ---
# Tic Tac Toe Player base class definition
# --------------------------------------------------------------------
# pylint: disable=too-few-public-methods

"""Implementation of the TttPlayer base abstract class:
    a generic player that has a piece and it is able to perform a move.
    This class shall not be used directly: it shall be used as a base
    class for real players
"""
from abc import ABC, abstractmethod

class TttPlayer(ABC):
    """A Tic Tac Toe player base class."""
    def __init__(self, piece, verbosity=0):
        """TttPlayer class constructor."""
        self.piece = piece
        self.__verbosity = verbosity
        if piece == "x":
            self.other_piece = "o"
        else:
            self.other_piece = "x"

        ABC.__init__(self)

    @abstractmethod
    def move(self, board):
        """Method to make a move. This is the base player,
            so this method is abstract."""

    def log_info(self, *args):
        """print a log line if verbosity > 0"""
        if self.__verbosity > 0:
            print(*args)



# --------------------------------------------------------------------
# jokettt project, v. 0.1
# by F. Piantini <francesco.piantini@gmail.com>
# ---
# Tic Tac Toe Board class definition
# --------------------------------------------------------------------
"""Implementation of the Board class: a board to play Tic Tac Toe game."""
import sys

from random import seed
from random import randint

import numpy as np

class Board:
    """A board to play Tic Tac Toe game."""
    # ------------------------------------------------------
    def __init__(self, first_piece, second_piece, init_board=None):

        """Board class constructor"""
        if init_board is not None:
            self.__board = init_board
        else:
            self.__board = [['_', '_', '_'],
                            ['_', '_', '_'],
                            ['_', '_', '_']]

        self.__first_piece = first_piece
        self.__second_piece = second_piece
        seed()
        self.__zobrist_hash = 0
        self.__init_zhash()

    # ------------------------------------------------------
    def reset(self, init_board=None):

        """Reset the board to the given schema (default = empty)."""
        if init_board is not None:
            for _x in range(0, 3):
                for _y in range(0, 3):
                    self.__board[_x][_y] = init_board[_x][_y]
        else:
            self.__board = [['_', '_', '_'],
                            ['_', '_', '_'],
                            ['_', '_', '_']]

        # initialize Zobrist hash value
        self.__evaluate_zhash()

    # ------------------------------------------------------
    def get_zhash(self):
        """Return the current Zobrist hash of the board."""
        return self.__zobrist_hash

    # ------------------------------------------------------
    def is_empty(self):
        """Returns True if the board is empty"""
        if self.__board == [['_', '_', '_'],
                            ['_', '_', '_'],
                            ['_', '_', '_']]:
            return True

        return False

    # ------------------------------------------------------
    def is_not_full(self):
        """Returns True if the board is not full."""
        for _x in range(0, 3):
            for _y in range(0, 3):
                if self.__board[_x][_y] == "_":
                    return True
        return False

    # ------------------------------------------------------
    def is_full(self):
        """Returns True if the board is full."""
        return not self.is_not_full()

    # ------------------------------------------------------
    def pos_is_empty(self, _x, _y):
        """Returns True if the given board position does not contains a pawn."""
        return bool(self.__board[_x][_y] == "_")

    # ------------------------------------------------------
    def pos_is_busy(self, _x, _y):
        """Returns True if the given board position contains a pawn."""
        return not self.pos_is_empty(_x, _y)

    # ------------------------------------------------------
    def valid_moves(self):
        """Returns the list of the valid moves in the current board state."""
        move_list = []
        for _x in range(0, 3):
            for _y in range(0, 3):
                if self.__board[_x][_y] == "_":
                    move_list.append([_x, _y])
        return move_list

    # ------------------------------------------------------
    def is_valid_move(self, move):
        """Returns True if the move is valid in the current board state."""
        # check the format of the move
        if len(move) != 2:
            return False
        _x, _y = self.convert_movestring_to_indexes(move)
        if _x == -1 or _y == -1:
            return False
        # check if the position if free in the board
        if self.pos_is_busy(_x, _y):
            return False
        return True

    # ------------------------------------------------------
    def analyze_move(self, move, piece):
        """analize a move, returning the new board hash,
           and the score of the position"""
        zhash, score = self.place_pawn(move[0], move[1], piece)
        # return the board in the previous status
        _ = self.__remove_pawn(move[0], move[1])
        return zhash, score

    # ------------------------------------------------------
    def place_pawn(self, _x, _y, piece):
        """Places a pawn in the given board position."""
        if self.pos_is_empty(_x, _y):
            self.__board[_x][_y] = piece
            self.__update_zhash(_x, _y, piece)
        return self.get_zhash(), self.evaluate(piece)

    # ------------------------------------------------------
    def evaluate(self, piece):
        """Evaluates the board value."""
        neg_piece = self.__get_other_piece(piece)
        score = self.__evaluate_rows(piece, neg_piece)
        if score != 0:
            return score
        score = self.__evaluate_cols(piece, neg_piece)
        if score != 0:
            return score
        return self.__evaluate_diags(piece, neg_piece)

    # ------------------------------------------------------
    def convert_movestring_to_indexes(self, move):
        """Convert the move from the <row><col> format (e.g. "A1")
        format to the board x,y indexes.
        """
        row = move[0].upper()
        col = move[1]
        return self.__convert_move_coords_to_indexes(row, col)

    # ------------------------------------------------------
    def convert_move_to_movestring(self, move):
        """Convert the move from the [x,y] move format
        to the <row><col> string format (e.g. "A1").
        """
        return self.__convert_indexes_to_movestring(move[0], move[1])

    # ------------------------------------------------------
    def __remove_pawn(self, _x, _y):
        """Removes a pawn from the given board position."""
        piece = self.__board[_x][_y]
        if piece != "_":
            self.__update_zhash(_x, _y, piece)
            self.__board[_x][_y] = "_"
        return piece

    # experimental code that try to explore the concept
    # of "equivalent boards"... could be used by learner
    # player to speed up learning. Temporarly disabled...
    # ------------------------------------------------------
    #def get_zhash_equivalent_boards(self):
    #    """Return the zhash of the current board and of all
    #        the equivant simmetrical boards"""
    #    zhash2 = self.__rotate_board_clockwise()
    #    zhash3 = self.__rotate_board_clockwise()
    #    zhash4 = self.__rotate_board_clockwise()
    #    zhash1 = self.__rotate_board_clockwise()
    #    return zhash1, zhash2, zhash3, zhash4

    # ------------------------------------------------------
    #def __replace_pawn(self, _x, _y, piece):
    #    """Replace a pawn in the given board position
    #        with the given piece."""
    #    old_piece = self.__remove_pawn(_x, _y)
    #    self.place_pawn(_x, _y, piece)
    #    return old_piece

    # ------------------------------------------------------
    #def __move_pawn(self, x0, y0, x1, y1):
    #    """Move the pawn in the [x0, y0] position to the
    #       [x1, y1] position. Returns the piece that was
    #       in the [x1, y1] position"""
    #    return self.__replace_pawn(x1, y1, self.__remove_pawn(x0, y0))

    #def __rotate_board_clockwise(self):
    #    """Build the board equivalent to the current one
    #        rotating it by 90 degrees clockwise"""
    #    piece = self.__board[0][0]
    #    piece = self.__replace_pawn(0, 2, piece)
    #    piece = self.__replace_pawn(2, 2, piece)
    #    piece = self.__replace_pawn(2, 0, piece)
    #    _ = self.__replace_pawn(0, 0, piece)
    #    piece = self.__board[0][1]
    #    piece = self.__replace_pawn(1, 2, piece)
    #    piece = self.__replace_pawn(2, 1, piece)
    #    piece = self.__replace_pawn(1, 0, piece)
    #    _ = self.__replace_pawn(0, 1, piece)
    #    return self.get_zhash()

    # ------------------------------------------------------
    @staticmethod
    def __convert_move_coords_to_indexes(row, col):
        """Convert move coordinates (e.g. "A","1") to board x,y indexes."""
        row_to_x = {
            "A": 0,
            "B": 1,
            "C": 2
        }
        col_to_y = {
            "1": 0,
            "2": 1,
            "3": 2
        }
        return row_to_x.get(row, -1), col_to_y.get(col, -1)

    # ------------------------------------------------------
    @staticmethod
    def __convert_indexes_to_movestring(_x, _y):
        """Convert the move from board x,y indexes to <row><col> format (e.g. "A1")."""
        x_to_row = {
            0: "A",
            1: "B",
            2: "C"
        }
        y_to_col = {
            0: "1",
            1: "2",
            2: "3"
        }
        mstring = ""
        mstring += x_to_row[_x]
        mstring += y_to_col[_y]
        return mstring

    # ------------------------------------------------------
    def __evaluate_rows(self, pos_piece, neg_piece):
        """Evaluates the board value checking only rows."""
        val = 0
        row = 0
        while val == 0 and row < 3:
            if self.__board[row][0] == self.__board[row][1] and \
                    self.__board[row][1] == self.__board[row][2]:
                if self.__board[row][0] == pos_piece:
                    val = 10
                elif self.__board[row][0] == neg_piece:
                    val = -10
            row += 1

        return val

    # ------------------------------------------------------
    def __evaluate_cols(self, pos_piece, neg_piece):
        """Evaluates the board value checking only columns."""
        val = 0
        col = 0
        while val == 0 and col < 3:
            if self.__board[0][col] == self.__board[1][col] and \
               self.__board[1][col] == self.__board[2][col]:
                if self.__board[0][col] == pos_piece:
                    val = 10
                elif self.__board[0][col] == neg_piece:
                    val = -10
            col += 1

        return val

    # ------------------------------------------------------
    def __evaluate_diags(self, pos_piece, neg_piece):
        """Evaluates the board value checking only diagonals."""
        val = 0
        if self.__board[0][0] == self.__board[1][1] and \
           self.__board[1][1] == self.__board[2][2]:
            if self.__board[1][1] == pos_piece:
                val = 10
            elif self.__board[1][1] == neg_piece:
                val = -10

        if val != 0:
            return val

        if self.__board[0][2] == self.__board[1][1] and \
           self.__board[1][1] == self.__board[2][0]:
            if self.__board[1][1] == pos_piece:
                val = 10
            elif self.__board[1][1] == neg_piece:
                val = -10

        return val

    # ------------------------------------------------------
    def __init_zhash(self):
        """Initialize Zobrist hash table."""
        # initialize Zobrist hash table
        self.__zhash_table = np.empty([3, 3, 2], dtype=int)

        for _x in range(0, 3):
            for _y in range(0, 3):
                for _e in range(0, 2):
                    self.__zhash_table[_x][_y][_e] = randint(0, sys.maxsize)
        # initialize Zobrist hash value
        self.__evaluate_zhash()

    # ------------------------------------------------------
    def __evaluate_zhash(self):
        """Initialize Zobrist hash value."""
        self.__zobrist_hash = 0
        for _x in range(0, 3):
            for _y in range(0, 3):
                piece = self.__board[_x][_y]
                if piece != "_":
                    piece_ndx = self.__convert_piece_in_index(piece)
                    self.__zobrist_hash ^= self.__zhash_table[_x][_y][piece_ndx]

    # ------------------------------------------------------
    def __update_zhash(self, _x, _y, piece):
        """Update Zobrist hash value after a board status change
        due to a single place or remove of a pawn.
        """
        piece_ndx = self.__convert_piece_in_index(piece)
        self.__zobrist_hash ^= self.__zhash_table[_x][_y][piece_ndx]

    # ------------------------------------------------------
    def __convert_piece_in_index(self, piece):
        """Convert a piece in internal index."""
        if piece == self.__first_piece:
            return 0
        return 1

    # ------------------------------------------------------
    def __get_other_piece(self, piece):
        if piece == self.__first_piece:
            return self.__second_piece
        return self.__first_piece

    # ------------------------------------------------------
    def __str__(self):
        """__str__ display of the board."""
        ###return '     1    2    3\nA %r\nB %r\nC %r\n--- hash = %r' % \
        ###    (self.__board[0], self.__board[1], self.__board[2], self.get_zhash())
        return '    1    2    3\nA %r\nB %r\nC %r\n' % \
            (self.__board[0], self.__board[1], self.__board[2])

    # ------------------------------------------------------
    def __repr__(self):
        """__repr__ representation of the board."""
        return 'Board(%s)' % self.__board



# --------------------------------------------------------------------
# jokettt project, v. 0.1
# by F. Piantini <francesco.piantini@gmail.com>
# ---
# Tic Tac Toe perfect minimax Player class definition
#
# --- Credits:
# https://www.geeksforgeeks.org/minimax-algorithm-in-game-theory-set-2-evaluation-function/
#
# ---
# Minimax algo applied to TicTacToe game
#  - with alpha-beta-pruning
#  - with (Zobrist) hash evaluation function
# --------------------------------------------------------------------

"""Implementation of the Minimax TttPlayer:
    a tic-tac-toe perfect minimax Player with alpha-beta-pruning
    This class is derived from the TttPlayer base class
    This player has a "dumb" mode that can be activated at any step:
    in this mode, a random move is chosen
"""
from copy import deepcopy

from random import randint
from random import shuffle

from tttplayer import TttPlayer

# ----------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------
# pylint: disable=too-few-public-methods
class MinimaxParameters:
    """Parameters for a minimax algorithm with alpha-beta pruning."""
    def __init__(self, depth=0, is_maximizer=True, alpha=-1000, beta=1000):
        self.depth = depth
        self.is_maximizer = is_maximizer
        self.alpha = alpha
        self.beta = beta
# pylint: enable=too-few-public-methods

# ----------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------
class TttMinimaxPlayer(TttPlayer):
    """A Tic Tac Toe minimax automatic player."""

    # ----------------------------------------------------------------------------------------
    def __init__(self, piece, dumb_mode=False, verbosity=0):
        """TttPlayer class constructor. Save the given piece,
            and enable dumb mode is requested."""
        TttPlayer.__init__(self, piece, verbosity)
        self.set_dumb_mode(dumb_mode)

    # ----------------------------------------------------------------------------------------
    def set_dumb_mode(self, dumb_mode):
        """Enable or disable the dumb mode"""
        self.__dumb_mode = dumb_mode

    # ----------------------------------------------------------------------------------------
    def move(self, board):
        """Do a move using currently selected mode (dumb or minimax)"""
        if self.__dumb_mode:
            return self.__move_dumb(board)
        return self.__move_smart(board)

    # ----------------------------------------------------------------------------------------
    def __move_smart(self, board):
        """Do a smart move (using minimax algo). If this is the first move,
            performs a random move using dumb mode"""
        if board.is_empty():
            ###print("I'm smart, but this is first move... choose random move")
            return self.__move_dumb(board)

        ###print("I'm smart, choose best move")
        _, best_x, best_y = self.__find_move_minimax(board, MinimaxParameters(0, True, -1000, 1000))
        return best_x, best_y

    # ----------------------------------------------------------------------------------------
    def __find_move_minimax(self, board, mm_par):
        """Find the best move (or one of the best) using the minimax algo"""
        best_x = None
        best_y = None
        val = board.evaluate(self.piece)
        if val != 0 or board.is_full():
            # evaluate function returns a positive value
            # if maximizer win, a negative value otherwise
            if val > 0:
                return val - mm_par.depth, best_x, best_y
            return val + mm_par.depth, best_x, best_y

        move_list = board.valid_moves()
        shuffle(move_list)  # to add some variability to the play (...maybe)
        if mm_par.is_maximizer:
            best_score = -1000
            for move in move_list:
                simul_board = deepcopy(board)
                simul_board.place_pawn(move[0], move[1], self.piece)
                score, _, _ = self.__find_move_minimax(simul_board, \
                       MinimaxParameters(mm_par.depth+1, False, mm_par.alpha, mm_par.beta))
                if score > best_score:
                    best_score = score
                    best_x = move[0]
                    best_y = move[1]
                mm_par.alpha = max(mm_par.alpha, best_score)
                if mm_par.beta <= mm_par.alpha:
                    break
        else:
            best_score = 1000
            for move in move_list:
                simul_board = deepcopy(board)
                simul_board.place_pawn(move[0], move[1], self.other_piece)
                score, _, _ = self.__find_move_minimax(simul_board, \
                       MinimaxParameters(mm_par.depth+1, True, mm_par.alpha, mm_par.beta))
                if score < best_score:
                    best_score = score
                    best_x = move[0]
                    best_y = move[1]
                mm_par.beta = min(mm_par.beta, best_score)
                if mm_par.beta <= mm_par.alpha:
                    break

        return best_score, best_x, best_y

    # ----------------------------------------------------------------------------------------
    @staticmethod
    def __move_dumb(board):
        """Do a dumb move"""
        ###print("I'm dumb, choose random move")
        if board.is_full():
            return None, None
        while True:
            _x = randint(0, 2)
            _y = randint(0, 2)
            if board.pos_is_empty(_x, _y):
                return _x, _y
